module LinRatExp

using LinearAlgebra

export BKstatic
export BlanKahn


function BKstatic(Jf::Matrix{Float64},Jg::Matrix{Float64},g::VecOrMat{Float64},cntrl::Vector{Int64},state::Vector{Int64},r::Real)
    n_cntrl = length(cntrl)
    n_state = length(state)
    n_vars = n_cntrl+n_state

    static_eq = findall(x->sum(abs.(Jf[x,:]))==0,1:n_vars)
    dynam_eq = findall(x->sum(abs.(Jf[x,:]))!=0,1:n_vars)
    n_static_eq = length(static_eq)
    n_dynam_eq = length(dynam_eq)

    if n_static_eq > n_cntrl
        error("There are more static equations than control variables")
    end

    if n_dynam_eq == n_state
        Theta = -Jg[static_eq,cntrl]\Jg[static_eq,state]
        theta = -Jg[static_eq,cntrl]\g[static_eq]
        Omega = (Jf[dynam_eq,state]+Jf[dynam_eq,cntrl]*Theta)\(Jg[dynam_eq,state]+Jg[dynam_eq,cntrl]*Theta)
        omega = (Jf[dynam_eq,state]+Jf[dynam_eq,cntrl]*Theta)\(g[dynam_eq]-Jf[dynam_eq,cntrl]*theta)
        policy = (Theta,theta,Omega,omega)
    elseif n_static_eq == 0
        M = Jf\Jg
        m = Jf\g
        policy = BlanKahn(M,m,cntrl,state,r)
    else
        cntrl_s = zeros(n_static_eq)
        cntrl_d = zeros(n_cntrl-n_static_eq)
        for i = 1:100
            k = rand(n_cntrl)
            K = sortperm(k)
            cntrl_s = cntrl[K[1:n_static_eq]]
            cntrl_d = cntrl[K[n_static_eq+1:n_cntrl]]
            if det(Jg[static_eq,cntrl_s]) != 0
                break
            end
        end

        Psi = -Jg[static_eq,cntrl_s]\[Jg[static_eq,cntrl_d] Jg[static_eq,state]]
        psi = -Jg[static_eq,cntrl_s]\g[static_eq]

        new_Jf = Jf[dynam_eq,cntrl_s]*Psi+[Jf[dynam_eq,cntrl_d] Jf[dynam_eq,state]]
        new_Jg = Jg[dynam_eq,cntrl_s]*Psi+[Jg[dynam_eq,cntrl_d] Jg[dynam_eq,state]]
        new_g = Jg[dynam_eq,cntrl_s]*psi+g[dynam_eq]


        new_cntrl = Vector(1:n_static_eq)
        new_state = Vector(n_static_eq+1:n_static_eq+n_state)

        new_M = new_Jf\new_Jg
        new_m = new_Jf\new_g
        new_policy = BlanKahn(new_M,new_m,new_cntrl,new_state,r) 

        new_Theta = new_policy[1]
        new_theta = new_policy[2]
        new_Omega = new_policy[3]
        new_omega = new_policy[4]


        Theta = zeros(n_cntrl,n_state)
        theta = zeros(n_cntrl)

        cs = findall(i->cntrl[i] ∈ cntrl_s,1:n_cntrl)
        cd = findall(i->cntrl[i] ∈ cntrl_d,1:n_cntrl)

        Theta[cs,:] = Psi*[new_Theta;I(n_state)]
        Theta[cd,:] = new_Theta        
        theta[cs] = psi
        theta[cd] = new_theta
        
        Omega = new_Omega
        omega = new_omega

        policy = (Theta,theta,Omega,omega)

    end
    return policy
end

function BlanKahn(M::Matrix{Float64},m::VecOrMat{Float64},cntrl::Vector{Int64},state::Vector{Int64},r::Real)
    n_cntrl = length(cntrl)
    n_state = length(state)

    L = eigvals(M)
    V = eigvecs(M)

    expl = findall(l-> l-(1+r)>=-1e-8,real(L))
    conv = findall(l-> l-(1+r)<-1e-8,real(L))

    n_expl = length(expl)
    if n_expl < n_cntrl
        error("There are infinite solutions.")
    elseif n_expl > n_cntrl
        error("There are no solutions.")
    end

    mm = V\m
    z_expl = (I(n_expl)-diagm(L[expl]))\mm[expl]
    #z_conv = V[state,conv]\(X0-V[state,expl]*z_expl)
    #Y0 = V[cntrl,conv]*z_conv+V[cntrl,expl]*z_expl
    Theta = V[cntrl,conv]/V[state,conv]
    theta = -Theta*V[state,expl]*z_expl+V[cntrl,expl]*z_expl

    L_conv = diagm(L[conv])
    L_expl = diagm(L[expl])
    #X1 = V[state,conv]*L_conv*z_conv+V[state,expl]*L_expl*z_expl+m[state]
    Omega = V[state,conv]*L_conv/V[state,conv]
    omega = -Omega*V[state,expl]*z_expl+V[state,expl]*L_expl*z_expl+m[state]

    policy = (Theta,theta,Omega,omega)
    return policy
end

end