module RiskSensiMethod

using SymPy
using PyCall
using LinearAlgebra

include("LinRatExp.jl")
using .LinRatExp

export derivModel
export derivModel2
export derivModel3
export derivVolat
export RiskSensi
export RiskSensiVol
export RiskSensi3

MM = Union{VecOrMat{Float64},Union{VecOrMat{Int64},VecOrMat{Real}}}
gg = Union{Vector{Float64},Union{Vector{Int64},Vector{Real}}}
SS = Union{VecOrMat{Sym{PyObject}},Sym{PyObject}}

function gsize(A::Matrix)
    n = size(A)
    return n
end

function gsize(A::LinearAlgebra.Adjoint)
    n = size(A)
    return n
end

function gsize(A::Union{Array{Float64, 3},Union{Array{Int64,3},Array{Real,3}}})
    n = size(A)
    return n
end

function gsize(A::Vector)
    l = length(A)
    n = (l,1)
    return n
end

function gsize(A::Real)
    n = (1,1)
    return n
end

function gsize(A::Sym{PyObject})
    n = (1,1)
    return n
end

function VecVar(M::Union{Real,MM})
    n1,n2 = gsize(M)
    if n1 != n2
        error("Dimension mismatch")
    end
    nn::Integer = n2*(n2+1)/2
    V = zeros(nn,1)
    c = 0
    for i = 1:n1
        for j = i:n2
            c = c+1
            V[c,1] = M[i,j]
        end
    end
    return V
end

function Gamma2(H::MM)
    n1,n2 = gsize(H)
    nn1::Integer = n1*(n1+1)/2
    nn2::Integer = n2*(n2+1)/2
    G = zeros(nn1,nn2)
    c = 0
    for j1 = 1:n2
        for j2 = j1:n2
            c = c+1
            G[:,c] = VecVar(H[:,j1]*H[:,j2]')+VecVar(H[:,j2]*H[:,j1]')
            if j1 == j2
                G[:,c] = G[:,c]/2
            end
        end
    end
    return G
end

function derivModel(F::Vector{Sym{PyObject}},G::Vector{Sym{PyObject}},vars::Tuple)
    n = length(F)
    m = length(G)
    nv = length(vars)
    if m != n
        error("Dimension mismatch")
    elseif n != nv
        error("Number of equations and variables must be the same")
    end
    JF = zeros(Sym,n,n)
    JG = similar(JF)
    for i = 1:n
        c = 0
        for j1 = 1:nv
            c = c+1
            JF[i,c] = diff(F[i],vars[j1])
            JG[i,c] = diff(G[i],vars[j1])
        end
    end
    function makefunc(h::Tuple)
        f::Matrix{Float64} = N(F.subs(vars.=>h))
        Jf::Matrix{Float64} = N(JF.subs(vars.=>h))
        g::Matrix{Float64} = N(G.subs(vars.=>h))
        Jg::Matrix{Float64} = N(JG.subs(vars.=>h))
        return f,Jf,g,Jg
    end
    return makefunc
end

function derivModel2(F::Vector{Sym{PyObject}},G::Vector{Sym{PyObject}},vars::Tuple)
    n = length(F)
    m = length(G)
    nv = length(vars)
    if m != n
        error("Dimension mismatch")
    elseif n != nv
        error("Number of equations and variables must be the same")
    end
    n2::Integer = n*(n+1)/2
    JF = zeros(Sym,n,n)
    JG = similar(JF)
    HF = zeros(Sym,n,n2)
    for i = 1:n
        c1 = c2 = 0
        for j1 = 1:nv
            c1 = c1+1
            JF[i,c1] = diff(F[i],vars[j1])
            JG[i,c1] = diff(G[i],vars[j1])
            for j2 = j1:nv
                c2 = c2+1
                HF[i,c2] = diff(F[i],vars[j1],vars[j2])
            end
        end
    end
    function makefunc(h::Tuple)
        f::Matrix{Float64} = N(F.subs(vars.=>h))
        Jf::Matrix{Float64} = N(JF.subs(vars.=>h))
        Hf::Matrix{Float64} = N(HF.subs(vars.=>h))
        g::Matrix{Float64} = N(G.subs(vars.=>h))
        Jg::Matrix{Float64} = N(JG.subs(vars.=>h))
        return f,Jf,Hf,g,Jg
    end
    return makefunc
end

function derivModel3(F::Vector{Sym{PyObject}},G::Vector{Sym{PyObject}},vars::Tuple)
    n = length(F)
    m = length(G)
    nv = length(vars)
    if m != n
        error("Dimension mismatch")
    elseif n != nv
        error("Number of equations and variables must be the same")
    end
    n2::Integer = n*(n+1)/2
    n3::Integer = n*(n+1)*(n+2)/6
    JF = zeros(Sym,n,n)
    JG = similar(JF)
    HF = zeros(Sym,n,n2)
    H3F = zeros(Sym,n,n3)
    for i = 1:n
        c1 = c2 = c3 = 0
        for j1 = 1:nv
            c1 = c1+1
            JF[i,c1] = diff(F[i],vars[j1])
            JG[i,c1] = diff(G[i],vars[j1])
            for j2 = j1:nv
                c2 = c2+1
                HF[i,c2] = diff(F[i],vars[j1],vars[j2])
                for j3 = j2:nv
                    c3 = c3+1
                    H3F[i,c3] = diff(F[i],vars[j1],vars[j2],vars[j3])
                end
            end
        end
    end
    function makefunc(h::Tuple)
        f::Matrix{Float64} = N(F.subs(vars.=>h))
        Jf::Matrix{Float64} = N(JF.subs(vars.=>h))
        Hf::Matrix{Float64} = N(HF.subs(vars.=>h))
        H3f::Matrix{Float64} = N(H3F.subs(vars.=>h))
        g::Matrix{Float64} = N(G.subs(vars.=>h))
        Jg::Matrix{Float64} = N(JG.subs(vars.=>h))
        return f,Jf,Hf,H3f,g,Jg
    end
    return makefunc
end

function derivVolat(S::SS,vars::Tuple)
    n1,n2 = gsize(S)
    n1 != n2 ? error("Matrix is not square") : 1
    n = length(vars)
    dS = zeros(Sym,n1,n2*n)
    for i = 1:n
        for j1 = 1:n1
            for j2 = 1:n2
                dS[j1,(i-1)*n2+j2] = diff(S[j1,j2],vars[i])
            end
        end
    end
    function makefunc(h)
        s::Union{MM,Real} = N.(S.subs(vars.=>h))
        ds::Union{MM,Real} = N.(dS.subs(vars.=>h))
        return s,ds
    end
    return makefunc 
end

function RiskSensi(F1::MM,F2::MM,G1::MM,g1::MM,S::Union{Real,MM},cntrl::Vector{Int64},state::Vector{Int64})
    ns = length(state)
    n = length(cntrl)+ns
    policy = BKstatic(F1,G1,g1,cntrl,state,0)
    Theta = policy[1]
    
    H = zeros(n,ns)
    o = ones(ns,n)
    H[state,:] = I(ns)
    H[cntrl,:] = Theta
    H2 = Gamma2(H).*Gamma2(o)'

    F = F1
    G = G1
    g = g1-1/2*F2*H2*VecVar(S)
 
    policy2 = BKstatic(F,G,g,cntrl,state,0)
    return policy2
end

function RiskSensiVol(F1::MM,F2::MM,G1::MM,g1::MM,S::Union{Real,MM},dS::Union{Real,MM},cntrl::Vector{Int64},state::Vector{Int64};tolv = 1e-7,maxiterv = 100,stepv = 1,showiterv = 0)
    ns = length(state)
    n = length(cntrl)+ns
    policy = BKstatic(F1,G1,g1,cntrl,state,0)
    
    nn::Integer = ns*(ns+1)/2
    vdS = zeros(nn,n)
    for i = 1:n
        vdS[:,i] = VecVar(dS[:,ns*(i-1)+1:ns*i])
    end

    policy2 = policy
    begin 
        local Theta = policy[1]
        o = ones(ns,n)
        for iter = 2:maxiterv
            H = zeros(n,ns)
            H[state,:] = I(ns)
            H[cntrl,:] = Theta
            H2 = Gamma2(H).*Gamma2(o)'

            F = F1
            G = G1-1/2*F2*H2*vdS
            g = g1-1/2*F2*H2*VecVar(S)
 
            policy2 = BKstatic(F,G,g,cntrl,state,0)
            nTheta = (1-stepv)*Theta+stepv*policy2[1]
            if maximum(abs.(nTheta-Theta)) < tolv
                showiterv == 1 ? println("Policy convergence in $iter iterations") : true
                break
            else
                Theta = nTheta
                if iter == maxiterv
                    println("Policy din't converge")
                end
            end
        end
    end
    return policy2
end

L = LinearAlgebra.Adjoint

function TripleMult(A::gg,B::L,C::L)
    nA = gsize(A)
    nB = gsize(B)
    nC = gsize(C)
    nA1 = nA[1]
    length(nA) == 1 ? nA2 = 1 : nA2 = nA[2]
    nB1 = nB[1]
    length(nB) == 1 ? nB2 = 1 : nB2 = nB[2]
    nC1 = nC[1]
    length(nC) == 1 ? nC2 = 1 : nC2 = nC[2]
    if nA2 != nB1
        error("dimension missmatch");
    elseif nA2 != nC1
        error("dimension of C");
    end
    f = zeros(nA1,nB2,nC2)
    for i1 = 1:nA1
        for i2 = 1:nB2
            for i3 = 1:nC2
                F = 0
                for j = 1:nA2
                    F = F+A[i1,j]*B[j,i2]*C[j,i3]
                end
                f[i1,i2,i3] = F
            end
        end
    end
    return f
end

function VecM3(cub::Union{Real,Array{Float64,3}})
    n = gsize(cub)
    n1 = n[1]
    if n1 == 1
        G = cub
    else
        aux::Matrix = cub[1,:,:]
        G = VecVar(aux)
        for i = 2:n1
            aux2::Matrix = cub[i,i:n1,i:n1]
            G = [G;VecVar(aux2)]
        end
    end
    f = G
    return f
end

function Gamma3(H::Union{Real,MM})
    n,m = gsize(H)
    k = 1
    mm::Integer = m*(m+1)*(m+2)/6
    nn::Integer = n*(n+1)*(n+2)/6
    f = zeros(nn,mm)
    for i1 = 1:m
      for i2 = i1:m
        for i3 = i2:m
            f[:,k] = VecM3(TripleMult(H[:,i1],H[:,i2]',H[:,i3]'))+VecM3(TripleMult(H[:,i1],H[:,i3]',H[:,i2]'))+
                VecM3(TripleMult(H[:,i2],H[:,i1]',H[:,i3]'))+VecM3(TripleMult(H[:,i2],H[:,i3]',H[:,i1]'))+
                    VecM3(TripleMult(H[:,i3],H[:,i1]',H[:,i2]'))+VecM3(TripleMult(H[:,i3],H[:,i2]',H[:,i1]'))
          if i1 != i2 && i2 == i3
            f[:,k] = f[:,k]/2;
          elseif i1 == i2 && i2 != i3
            f[:,k] = f[:,k]/2;
          elseif i1 == i2 && i2 == i3
            f[:,k] = f[:,k]/6;
          end
          k = k+1;
        end
      end
    end
    return f
end

function RiskSensi3(F1::MM,F2::MM,F3::MM,G1::MM,g1::MM,Sig::Union{Real,MM},Xi::Union{Real,Array{Float64,3}},cntrl::Vector{Int64},state::Vector{Int64})
    nc = length(cntrl)
    ns = length(state)
    n = nc+ns
    #ns3::Integer = ns*(ns+1)*(ns+2)/6
  
    policy = BKstatic(F1,G1,g1,cntrl,state,0)
    Theta = policy[1]

    H = zeros(n,ns)
    o = ones(ns,n)
    H[state,:] = I(ns)
    H[cntrl,:] = Theta
    H2 = Gamma2(H).*Gamma2(o)'
    H3 = Gamma3(H).*Gamma3(o)'
 
    F = F1
    G = G1
    g = g1-1/2*F2*H2*VecVar(Sig)-1/6*F3*H3*VecM3(Xi)  
    
    policy2 = BKstatic(F,G,g,cntrl,state,0)
    return policy2
end

end