% This program computes a risk sensitive linear approximation to the
% price-dividend ratio of the asset pricing model of de Groot (2015, JEDC) 
% used in ``Risk sensitive linear approximations,''
% by Gustavo Solorzano Andrade and Juan Carlos Parra-Alvarez. 
%
% Important: Add folder utils to the path
%
% User inputs:
% Input 1: Parameter values 
% Input 2: Symbolic definition of control and state variables
% Input 3: Define the Euler system
% Input 4: Deterministic steady state (DSS)
%
% (c) Gustavo Solorzano Andrade and Juan Carlos Parra-Alvarez
% Date February 2, 2024.


%% ====================================================================== %
%                             Housekeeping                                %
% ----------------------------------------------------------------------- %
%close all
clear
clc

%% ====================================================================== %
%                    Add working directory and toolbox                    %
% ----------------------------------------------------------------------- %
addpath('../utils/');

%% ====================================================================== %
%                              Parameters                                 %
% ----------------------------------------------------------------------- %
beta   = 0.95;
gam    = 2.5;
z_star = 0.0179;
phi_z  = -0.139;
x_star = 0.0012;
phi_x  = 0.8552;
w      = 0.74e-5;

% Parametric constraint for convergence of analytical solution (Thm 3, pp. 313) 
cons = beta*exp((1-gam)*z_star+0.5*((1-gam)/(1-phi_z))^2*x_star...
    +((1-gam)^4)/(8*(1-phi_z)^4*(1-phi_x)^2)*w^2);
if cons >= 1
    error('Analytical price-dividend ratio does not converge: check parameter values')
end

% Algorithm options
randn('seed',305);
shock       = 1; %z stn dev (1), x stn dev (2), random (3)
Tsim        = 21; 
plots       = 1;
maxiter     = 100000;
crit        = sqrt(eps);
order       = 2;

%% =======================================================================%
%                                 Input 2:                                %
%                       Define control and states (y,x)                   %
%-------------------------------------------------------------------------%
syms lq z x
h       = [lq z x];
cntrl   = [1];
state   = [2 3];
n_state = length(state);
n       = length(h);

%% =======================================================================%
%                                 Input 3:                                %
%                           Model's definition                            %
% ----------------------------------------------------------------------- %
F = [beta*exp((1-gam)*z)*(1+exp(lq));
     z;
     x];

G = [exp(lq);
     (1-phi_z)*z_star+phi_z*z;
     (1-phi_x)*x_star+phi_x*x];
 
%% =======================================================================%
%                                 Input 4:                                %
%                     Deterministic steady state (DSS)                    %
%-------------------------------------------------------------------------%
var_ss = solve(F-G,h);
lq_ss   = var_ss.lq; 
z_ss   = var_ss.z; 
x_ss   = var_ss.x; 

%% ====================================================================== %
%                    Build model's solution approximation                 %
% ----------------------------------------------------------------------- %
[Jf,Hf,H3f,Jg,Hg,H3g] = derivModel(F,G,h,order);

Sig      = sym(zeros(n_state));
Sig(1,1) = (1-phi_x)*x_star+phi_x*x;
Sig(2,2) = w^2;
DSig     = derivVolat(Sig,h);

lq = eval(lq_ss);
z  = eval(z_ss);
x  = eval(x_ss);
for adjust = 1:maxiter
  nF  = eval(F);
  nG  = eval(G);
  nJf = eval(Jf);
  nJg = eval(Jg);
  if order ~= 1
    nHf = eval(Hf);
  end
  
  nSig  = eval(Sig);
  nDSig = eval(DSig); 
  
  if order == 1
      [Theta,theta,Omega,omega] = ParrSolv(nJf,0,nJg,nG-nF,nSig,nDSig,cntrl,state);
  else
      [Theta,theta,Omega,omega] = ParrSolv(nJf,nHf,nJg,nG-nF,nSig,nDSig,cntrl,state);
  end

  SS(state,1) = (eye(n_state)-Omega)\omega;
  SS(cntrl,1) = Theta*SS(state,1)+theta;
  if max(abs(SS))<crit
    break
  else
    lq = lq+SS(1);
    z = z+SS(2);
    x = x+SS(3);
  end
end

%% ====================================================================== %
%                        Stochastic Steady state                          %
% ----------------------------------------------------------------------- %
var_sss = [lq;z;x];

%% ====================================================================== %
%                           Model simulation                              %
% ----------------------------------------------------------------------- %
if shock == 1
  eps_z = zeros(Tsim,1);
  eps_z(2) = 1;
  eps_x = zeros(Tsim,1);
elseif shock == 2
  eps_x = zeros(Tsim,1);
  eps_x(2) = 1;
  eps_z = zeros(Tsim,1);
elseif shock == 3
  eps_z = randn(Tsim,1);
  eps_x = randn(Tsim,1);
else
  warning('No shock selected');
  eps_z = zeros(Tsim,1);
  eps_x = zeros(Tsim,1);
end
v = chol(nSig)';

%% ====================================================================== %
%                          Policy functions                               %
% ----------------------------------------------------------------------- %
Model_II_Analytical;
xx(1) = x+w*eps_x(1);
zz(1) = z+sqrt(xx(1))*eps_z(1);
X0    = [0;0];
for t = 1:Tsim
  % Approximated solution  
  X0 = X0+v*[eps_z(t);eps_x(t)];
  Y0 = Theta*X0+theta;
  variables(t,cntrl) = Y0';
  variables(t,state) = X0';
  X0 = Omega*X0+omega;  
  
  % Closed form solution
  zx     = [zz(t),xx(t)];
  qq(t)  = f_analytical_de_groot(zx);
  if t < Tsim
    xx(t+1) = (1-phi_x)*x_star+phi_x*xx(t)+w*eps_x(t+1);
    zz(t+1) = (1-phi_z)*z_star+phi_z*zz(t)+sqrt(xx(t+1))*eps_z(t+1);
  end
end

%% ====================================================================== %
%                                 Plots                                   %
% ----------------------------------------------------------------------- %
if plots == 1
    t      = zeros(Tsim,1);
    t(1,1) = 0;
    for i = 2:Tsim
        t(i,1) = t(i-1,1) + 1;
    end
    
    fig = figure(1);    
    subplot(3,1,1)
    plot(t(1:end),exp(variables(:,1)+lq),'-','linewidth',2);    
    title('Price/dividend ratio','interpreter','latex');
    axis([0 Tsim-1 -inf inf]);
    grid on;
    set(gca,'fontsize',15, 'box', 'off')    
    set(gcf,'color','w')  
    
    subplot(3,1,2)
    plot(t(1:end),variables(:,2)+z,'-','linewidth',2);  
    title('Dividend growth','interpreter','latex');
    axis([0 Tsim-1 -inf inf]);
    grid on;
    set(gca,'fontsize',15, 'box', 'off')    
    set(gcf,'color','w')  
    
    subplot(3,1,3)
    plot(t(1:end),variables(:,3)+x,'-','linewidth',2);    
    title('Conditional variance of dividend growth','interpreter','latex');
    axis([0 Tsim-1 -inf inf]);
    grid on;
    set(gca,'fontsize',15, 'box', 'off')    
    set(gcf,'color','w')  
    h2 = legend('Risk Sensitive linearization','True solution','Location','NorthEast');
    set(h2, 'interpreter','latex','fontsize',15)
    legend('boxoff')

    %% Plot policy function in Figure 1 of the paper
    grayColor = [.7 .7 .7];
    z_ss = z_star;
    x_ss = x_star;
    q_ss = (beta*exp((1-gam)*z_star))/(1-beta*exp((1-gam)*z_star));
    v_ss = 0;

    n_points = 50;
    z_range = linspace(-0.25,0.25,n_points);
    z_range = sort([z_range,z_star]);
    x_range = linspace(0.000001,4*x_ss,n_points);
    x_range = sort([x_range,x_star]);
    v_range = linspace(-8,7,n_points);
    v_range = sort([v_range,v_ss]);

    % Allocation of memory
    q_order2 = NaN(length(z_range),length(x_range));
    z_order2  = NaN(length(z_range),length(x_range));
    x_order2  = NaN(length(z_range),length(x_range));

    % Approximated stochastic steady state
    lq_order2_sss = var_sss(1);
    z_order2_sss = var_sss(2);
    x_order2_sss = var_sss(3);

    for i = 1:length(z_range)
        for j = 1:length(x_range)
            lq_order2(i,j) = Theta*[z_range(i)-z_order2_sss;x_range(j)-x_order2_sss]+theta;
            z_order2(i,j) = Omega(1,:)*[z_range(i)-z_order2_sss;x_range(j)-x_order2_sss]+omega(1,1);
            x_order2(i,j) = Omega(2,:)*[z_range(i)-z_order2_sss;x_range(j)-x_order2_sss]+omega(2,1);
        end
    end
    q_order2 = exp(lq_order2 + lq_order2_sss);
    z_order2 = z_order2 + z_order2_sss;
    x_order2 = x_order2 + x_order2_sss;

    Model_II_Analytical;
    q_True = NaN(length(z_range),length(x_range));
    for i = 1:length(z_range)
        for j =1:length(x_range)
            q_True(i,j) = f_analytical_de_groot([z_range(i);x_range(j)]);
        end
    end

    fig = figure(2);    
    subplot(1,2,1)
    plot(z_range-z_ss,q_order2(:,14),'r','linewidth',2.5);
    hold on
    plot(z_range-z_ss,q_True(:,14),'b--','linewidth',1.5);
    hold on
    plot(0,q_ss,'Marker','o','MarkerSize',8,'MarkerEdgeColor','k','LineWidth',1.5)
    hold on
    plot(0,exp(lq_order2_sss),'Marker','o','MarkerSize',8,'MarkerEdgeColor','r','LineWidth',1.5)
    hold off
    ylabel('Model II','interpreter','latex')
    xlabel('$z_t-\bar{z}$ $(x_t = \bar{x})$','interpreter','latex')
    grid on;
    box off
    axis([-inf inf -inf inf])
    h2 = legend('Risk-Sensitive','True','Location','SouthEast');
    set(h2,'FontSize',14,'interpreter','latex')
    legend('boxoff') 
    set(gca,'fontsize',16,'box','off')
    set(gcf, 'color','w')

    subplot(1,2,2)
    plot(x_range-x_ss,q_order2(28,:),'r','linewidth',2.5);
    hold on
    plot(x_range-x_ss,q_True(28,:),'b--','linewidth',1.5);
    hold on
    plot(0,q_ss,'Marker','o','MarkerSize',8,'MarkerEdgeColor','k','LineWidth',1.5)
    hold on
    plot(0,exp(lq_order2_sss),'Marker','o','MarkerSize',8,'MarkerEdgeColor','r','LineWidth',1.5)
    hold off
    xlabel('$x_t-\bar{x}$ $(z_t = \bar{z})$','interpreter','latex')
    grid on
    set(gca,'fontsize',16,'box','off')
    set(gcf, 'color','w')
    axis([-inf inf -inf inf])
end

% end-of-file